(module sqlite mzscheme
  (require (lib "etc.ss")
           (lib "contract.ss") 
           (lib "sqlite-ffi.ss" "sqlite"))
  (provide (all-defined))
  
  ; Struct
  (define-struct db (_db_ptr _errMsg_ptr _callback_ptr))
  (define/contract db-handle
    (db? . -> . cpointer?)
    (lambda (db)
      (ptr-ref (db-_db_ptr db) _sqlite3_ptr)))
  (define-struct statement (_stmt_ptr))
  (define/contract statement-handle
    (statement? . -> . cpointer?)
    (lambda (stmt)
      (ptr-ref (statement-_stmt_ptr stmt) _sqlite3_stmt_ptr)))
  
  ; Helpers
  (define/contract list->equal-sublists
    (integer? list? . -> . (listof list?))
    (lambda (n l)
      (let loop ([l l] [i 0] [r `()])
        (if (null? l)
            (reverse r)
            (loop (cdr l) (modulo (+ i 1) n)
                  (if (= i 0)
                      (append (list (list (car l)))
                              r)
                      (append (list (append (car r) (list (car l))))
                              (cdr r))))))))
  (define/contract sqlite-ok?
    (integer? . -> . boolean?)
    (lambda (s)
      (= s SQLITE_OK)))
  
  ; Methods
  (define/contract open
    (string? . -> . db?)
    (lambda (db-path)
      (let ([r (make-db (malloc _sqlite3_ptr_ptr)
                        (malloc _string_ptr)
                        (malloc _pointer))])
        (if (sqlite-ok? (sqlite3_open db-path (db-_db_ptr r)))
            r
            (begin (close r)
                   #f)))))
  (define/contract close
    (db? . -> . integer?)
    (lambda (db)
      (let ([s (sqlite3_close (db-handle db))])
        (when (sqlite-ok? s)
          (free (db-_db_ptr db))
          (free (db-_errMsg_ptr db))
          (free (db-_callback_ptr db))
          (set-db-_db_ptr! db #f)
          (set-db-_errMsg_ptr! db #f)
          (set-db-_callback_ptr! db #f))
        s)))
  
  (define/contract exec
    (db? string? (list? list? . -> . integer?) . -> . integer?)
    (lambda (db sql callback)
      (sqlite3_exec (db-handle db)
                    sql
                    (contract
                     (cpointer? integer? cpointer? cpointer? . -> . integer?)
                     (lambda (arg_ptr column-count_int column-values_ptr column-names_ptr)
                       (callback
                        (cvector->list (make-cvector* column-names_ptr _string column-count_int))
                        (cvector->list (make-cvector* column-values_ptr _string column-count_int))))
                     'positive 'negative)
                    (db-_callback_ptr db)
                    (db-_errMsg_ptr db))))
  (define/contract select
    (db? string? . -> . (listof list?))
    (lambda (db sql)
      (let ([result_ptr_ptr (malloc _string_array_ptr)]
            [row-count_ptr (malloc _int_ptr)]
            [column-count_ptr (malloc _int_ptr)])
        (let ([r (if (sqlite3_get_table (db-handle db)
                                        sql
                                        result_ptr_ptr
                                        row-count_ptr
                                        column-count_ptr
                                        (db-_errMsg_ptr db))
                     (list->equal-sublists (ptr-ref column-count_ptr _int)
                                           (cvector->list (make-cvector* (ptr-ref result_ptr_ptr _string_array) _string 
                                                                         (* (+ (ptr-ref row-count_ptr _int) 1)
                                                                            (ptr-ref column-count_ptr _int)))))
                     #f)])
          (sqlite3_free_table (ptr-ref result_ptr_ptr _string_array))
          (free result_ptr_ptr)
          (free row-count_ptr)
          (free column-count_ptr)
          r))))
  
  (define/contract prepare
    (db? string? . -> . statement?)
    (lambda (db sql)
      (let ([stmt (make-statement 
                   (malloc _sqlite3_stmt_ptr_ptr))])
        (if (sqlite-ok? (sqlite3_prepare (db-handle db)
                                         sql
                                         (string-length sql)
                                         (statement-_stmt_ptr stmt)
                                         (db-_errMsg_ptr db)))
            stmt
            (begin (finalize stmt)
                   #f)))))
  (define/contract run
    ((statement?) list? . ->* . (integer?))
    (lambda (stmt . params)
      (when (sqlite-ok? (sqlite3_reset (statement-handle stmt)))
        (let ([SQLITE_TRANSIENT/sleaziness -1])
          (for-each
           (lambda (i)
             (sqlite3_bind_text (statement-handle stmt) (+ i 1)
                                (list-ref params i)
                                (string-length (list-ref params i))
                                SQLITE_TRANSIENT/sleaziness))
           (build-list (sqlite3_bind_parameter_count (statement-handle stmt)) identity)))
        (sqlite3_step (statement-handle stmt)))))
  (define/contract finalize
    (statement? . -> . integer?)
    (lambda (stmt)
      (let ([s (sqlite3_finalize (statement-handle stmt))])
        (free (statement-_stmt_ptr stmt))
        (set-statement-_stmt_ptr! stmt #f)
        s)))
  
  (define/contract changes-count
    (db? . -> . integer?)
    (lambda (db)
      (sqlite3_changes (db-handle db))))
  (define/contract total-changes-count
    (db? . -> . integer?)
    (lambda (db)
      (sqlite3_total_changes (db-handle db)))))
